from tidypath import savedata, savefig, add_arg, modify_arg, delete_arg
from . import np_utils
from . import pd_utils
from . import storage
from . import plots
from . import stats
from .stats.rtopy import resample
from .stats.test import permutation
